#include "stdafx.h"

/*
    MsgTableEditor: Easily edit an extracted message table
    Copyright (C) 2010  Jeffrey Bush <jeff@coderforlife.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/gpl.html>.
*/

/*


* means done
! means done but not tested


* Q Quit (EOF also works)
* H Help (? also works)
* D Display Message Table

* S Save
* A Save As
* L Load
* N New File

* R Replace Message
  M Add Message
  B Add Block
  X Delete Message
  W Delete Block
! I Set ID Base

*/

#define MAX_WORD		65535
#define MAX_DWORD		18446744073709551615
#define MAX_TEXT_LEN	((MAX_WORD-2*sizeof(WORD))/sizeof(TCHAR))

#ifdef _UNICODE
#define UNI	TRUE
#else
#define UNI FALSE
#endif

typedef struct _MT_ENTRY {
	DWORD id;
	BOOL unicode;
	union {
		WCHAR *wText;
		CHAR *aText;
		TCHAR *tText;
	};
} MT_ENTRY;

typedef struct _MT_BLOCK {
	DWORD lowId;
	DWORD highId;
	DWORD nEntries;
	MT_ENTRY **entries;
} MT_BLOCK;

typedef struct _MT_DATA {
	TCHAR *file;
	DWORD nBlocks;
	MT_BLOCK **blocks;
} MT_DATA;

static WCHAR *CopyStrW(WCHAR *s) { return wcscpy(LocalAlloc(0, sizeof(WCHAR)*(wcslen(s)+1)), s); }
static CHAR  *CopyStrA(CHAR  *s) { return strcpy(LocalAlloc(0, sizeof(CHAR )*(strlen(s)+1)), s); }
#ifdef _UNICODE
#define CopyStr(s) CopyStrW(s)
#else
#define CopyStr(s) CopyStrA(s)
#endif

static WCHAR *GetStringW(WCHAR *s) {
	size_t i = wcslen(s)+1;
	BOOL end = TRUE;
	WCHAR *x = LocalAlloc(0, sizeof(WCHAR)*i);
	do {
		--i;
		x[i] = (end && (end = (s[i] == 0 || s[i] == L'\n' || s[i] == L'\r'))) ? 0: s[i];
	} while (i);
	return x;
}

static CHAR *GetStringA(CHAR *s) {
	size_t i = strlen(s)+1;
	BOOL end = TRUE;
	CHAR *x = LocalAlloc(0, sizeof(CHAR)*i);
	do {
		--i;
		x[i] = (end && (end = (s[i] == 0 || s[i] == '\n' || s[i] == '\r'))) ? 0: s[i];
	} while (i);
	return x;
}

static MT_DATA *CreateData(TCHAR *file, DWORD nBlocks) {
	MT_DATA *d = LocalAlloc(0, sizeof(MT_DATA));
	d->file = file;
	d->nBlocks = nBlocks;
	d->blocks = LocalAlloc(LMEM_ZEROINIT, sizeof(MT_BLOCK*)*nBlocks);
	return d;
}

static MT_BLOCK *CreateBlock(DWORD lowId, DWORD highId) {
	MT_BLOCK *b = LocalAlloc(0, sizeof(MT_BLOCK));
	b->lowId = lowId;
	b->highId = highId;
	b->nEntries = highId-lowId+1;
	b->entries = LocalAlloc(LMEM_ZEROINIT, sizeof(MT_ENTRY*)*b->nEntries);
	return b;
}

static MT_ENTRY *CreateEntry(DWORD id, BOOL unicode) {
	MT_ENTRY *e = LocalAlloc(0, sizeof(MT_ENTRY));
	e->id = id;
	e->unicode = unicode;
	e->aText = NULL;
	return e;
}

static MT_DATA *BuildData(TCHAR *file, BYTE *x) {
	DWORD i, j, offset;
	MESSAGE_RESOURCE_DATA *data = (MESSAGE_RESOURCE_DATA*)x;

	MT_DATA *d = CreateData(file, data->NumberOfBlocks);

	for (i = 0; i < d->nBlocks; ++i) {
		MESSAGE_RESOURCE_BLOCK *block = data->Blocks+i;
		d->blocks[i] = CreateBlock(block->LowId, block->HighId);
		offset = block->OffsetToEntries;
		for (j = 0; j < d->blocks[i]->nEntries; ++j) {
			MESSAGE_RESOURCE_ENTRY *entry = (MESSAGE_RESOURCE_ENTRY*)(x+offset);
			d->blocks[i]->entries[j] = CreateEntry(j+block->LowId, (entry->Flags & MESSAGE_RESOURCE_UNICODE) == MESSAGE_RESOURCE_UNICODE);
			if (d->blocks[i]->entries[j]->unicode) {
				d->blocks[i]->entries[j]->wText = GetStringW((WCHAR*)entry->Text);
			} else {
				d->blocks[i]->entries[j]->aText = GetStringA((CHAR*)entry->Text);
			}
			offset += entry->Length;
		}
	}
	
	return d;
}

static MT_DATA *ReadMessageTable(TCHAR *file) {
	DWORD size, read;
	BYTE *data = NULL;
	HANDLE f = CreateFile(file, FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (f == INVALID_HANDLE_VALUE) {
		_ftprintf(stderr, TEXT("! Could not open file '%s': %u\n"), file, GetLastError());
		return NULL;
	}
	size = GetFileSize(f, NULL);
	if (size == INVALID_FILE_SIZE) {
		_ftprintf(stderr, TEXT("! GetFileSize: %u\n"), GetLastError());
	} else if (!(data = LocalAlloc(0, size))) {
		_ftprintf(stderr, TEXT("! LocalAlloc: %u\n"), GetLastError());
	} else if (!ReadFile(f, data, size, &read, NULL)) {
		_ftprintf(stderr, TEXT("! ReadFile: %u\n"), GetLastError());
		data = LocalFree(data);
	} else if (read != size) {
		_ftprintf(stderr, TEXT("! Didn't read all bytes: %u!=%u\n"), read, size);
		data = LocalFree(data);
	}
	CloseHandle(f);
	if (data) {
		MT_DATA *x = BuildData(file, data);
		LocalFree(data);
		return x;
	} 
	return NULL;
}

static MT_DATA *CreateBlankMessageTable() {
	MT_DATA *data = CreateData(NULL, 1);
	data->blocks[0] = CreateBlock(1, 1);
	data->blocks[0]->entries[0] = CreateEntry(1, TRUE);
	data->blocks[0]->entries[0]->tText = CopyStr(TEXT("Example Message"));
	return data;
}

static TCHAR *ReadLine(TCHAR *s, int size) {
	int i;
	if (_fgetts(s, size, stdin) == NULL) {
		return NULL;
	}
	for (i = _tcslen(s)-1; i > 1 && (s[i] == TEXT('\r') || s[i] == TEXT('\n')); --i) {
		s[i] = 0;
	}
	return s;
}

static int ReadChar() {
	CHAR temp[100];
	if (fgets(temp, 100, stdin) == NULL) {
		return EOF;
	}
	return (int)temp[0];
}

static WORD align4(WORD x) {
	return x + x % 4;
}

static BYTE *CompileData(MT_DATA *d, DWORD *size) {
	DWORD i, j, sz, off = d->nBlocks*sizeof(MESSAGE_RESOURCE_BLOCK)+sizeof(DWORD);
	BYTE *bytes;
	MESSAGE_RESOURCE_DATA *data;

	sz = off;
	for (i = 0; i < d->nBlocks; ++i) {
		sz += 2*d->blocks[i]->nEntries*sizeof(WORD);
		for (j = 0; j < d->blocks[i]->nEntries; ++j) {
			MT_ENTRY *e = d->blocks[i]->entries[j];
			if (e->unicode) {
				sz += align4(sizeof(WCHAR)*(wcslen(e->wText)+3)); // + 3 for 0D 0A 00
			} else {
				sz += align4(sizeof(CHAR )*(strlen(e->aText)+3));
			}
		}
	}
	*size = sz;
	bytes = LocalAlloc(LMEM_ZEROINIT, sz);
	data = (MESSAGE_RESOURCE_DATA*)bytes;
	data->NumberOfBlocks = d->nBlocks;
	for (i = 0; i < d->nBlocks; ++i) {
		data->Blocks[i].LowId = d->blocks[i]->lowId;
		data->Blocks[i].HighId = d->blocks[i]->highId;
		data->Blocks[i].OffsetToEntries = off;
		for (j = 0; j < d->blocks[i]->nEntries; ++j) {
			MESSAGE_RESOURCE_ENTRY *entry = (MESSAGE_RESOURCE_ENTRY*)(bytes+off);
			MT_ENTRY *e = d->blocks[i]->entries[j];
			WORD sz;
			if (e->unicode) {
				sz = (wcslen(e->wText)+3)*sizeof(WCHAR);
				entry->Flags = MESSAGE_RESOURCE_UNICODE;
				wcscat(wcscpy((WCHAR*)entry->Text, e->wText), L"\r\n\0");
			} else {
				sz = (strlen(e->aText)+3)*sizeof(CHAR);
				entry->Flags = 0;
				strcat(strcpy((CHAR *)entry->Text, e->aText), "\r\n\0");
			}
			entry->Length = align4(sz)+2*sizeof(WORD);
			off += entry->Length;
		}
	}

	return bytes;
}

static BOOL SaveMessageTable(MT_DATA *data, TCHAR *file) {
	BYTE *bytes;
	DWORD size, wrote, retval = FALSE;
	HANDLE f;
	if (file==NULL) {
		file = data->file;
	} else {
		data->file = CopyStr(file);
	}
	bytes = CompileData(data, &size);
	f = CreateFile(file, FILE_GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (f == INVALID_HANDLE_VALUE) {
		_ftprintf(stderr, TEXT("! Could not open file '%s': %u\n"), file, GetLastError());
	} else if (!WriteFile(f, bytes, size, &wrote, NULL)) {
		_ftprintf(stderr, TEXT("! WriteFile: %u\n"), GetLastError());
	} else if (wrote != size) {
		_ftprintf(stderr, TEXT("! Didn't write all bytes: %u!=%u\n"), wrote, size);
	} else {
		retval = TRUE;
	}
	LocalFree(bytes);
	CloseHandle(f);
	if (retval)
		_tprintf(TEXT("Successfully saved to %s\n"), file);
	return retval;
}

static BOOL AskToSave(MT_DATA *data) {
	_tprintf(TEXT("Current data not saved, do you wish to save now? [Y/N/X] "));
	while (TRUE) {
		int c = toupper(ReadChar());
		if (c == 'Y') {
			return SaveMessageTable(data, NULL);
		} else if (c == 'N') {
			return TRUE;
		} else if (c == 'X') {
			return FALSE;
		} else {
			_tprintf(TEXT("Please enter Y to save, N to not save, or X to cancel: "));
		}
	}
}

static MT_ENTRY *GetEntry(MT_BLOCK *block, DWORD id) {
	return (id < block->lowId || id > block->highId) ? NULL : block->entries[id-block->lowId];
}

static MT_ENTRY *DestroyEntry(MT_ENTRY *entry) {
	entry->wText = LocalFree(entry->wText);
	return LocalFree(entry);
}

static MT_BLOCK *DestroyBlock(MT_BLOCK *block) {
	DWORD i;
	for (i = 0; i < block->nEntries; ++i) {
		block->entries[i] = DestroyEntry(block->entries[i]);
	}
	block->entries = LocalFree(block->entries);
	return LocalFree(block);
}

static MT_DATA *DestroyData(MT_DATA *data) {
	DWORD i;
	for (i = 0; i < data->nBlocks; ++i) {
		data->blocks[i] = DestroyBlock(data->blocks[i]);
	}
	data->blocks = LocalFree(data->blocks);
	data->file = LocalFree(data->file);
	return LocalFree(data);
}

static void PrintText(MT_ENTRY *entry) {
	if (entry->unicode) {
		wprintf(L"\"%s\"\n", entry->wText);
	} else {
		printf("\"%s\"\n", entry->aText);
	}
}

static void DisplayMessageTable(MT_DATA *data) {
	DWORD i, j;
	_tprintf(TEXT("File: %s\n"), data->file);
	for (i = 0; i < data->nBlocks; ++i) {
		_tprintf(TEXT("Block: %u\n"), i);
		for (j = 0; j < data->blocks[i]->nEntries; ++j) {
			wprintf(L"  %u: ", data->blocks[i]->entries[j]->id);
			PrintText(data->blocks[i]->entries[j]);
		}
	}
}

static DWORD AskBlockId(MT_DATA *data) {
	DWORD b_id;
	if (data->nBlocks == 1) {
		_tprintf(TEXT("Selected Block 0 (only 1 block to select)\n"));
		b_id = 0;
	} else {
		_tprintf(TEXT("Block (0-%u): "), data->nBlocks-1);
		_tscanf(TEXT("%u"), &b_id);
		while (b_id >= data->nBlocks) {
			_tprintf(TEXT("Invalid Block, valid choices are 0-%u: "), data->nBlocks-1);
			_tscanf(TEXT("%u"), &b_id);
		}
		ReadChar(); // clears the new line character
	}
	return b_id;
}

static DWORD AskEntryId(MT_BLOCK *block) {
	DWORD e_id;
	if (block->nEntries == 1) {
		_tprintf(TEXT("Selected Msg ID %u (only 1 msg id to select)\n"), block->lowId);
		e_id = block->lowId;
	} else {
		_tprintf(TEXT("Msg Id (%u-%u): "), block->lowId, block->highId);
		_tscanf(TEXT("%u"), &e_id);
		while (e_id < block->lowId || e_id > block->highId) {
			_tprintf(TEXT("Invalid Msg ID, valid choices are %u-%u: "), block->lowId, block->highId);
			_tscanf(TEXT("%u"), &e_id);
		}
		ReadChar(); // clears the new line character
	}
	return e_id;
}

static void UpdateEntryIds(MT_BLOCK *block) {
	DWORD id, i;
	block->highId = block->lowId+block->nEntries-1;
	for (id = block->lowId, i = 0; i <= block->nEntries; ++id, ++i) {
		block->entries[i]->id = id;
	}
}

static void ReplaceMessage(MT_DATA *data) {
	DWORD b_id = AskBlockId(data);
	MT_BLOCK *block = data->blocks[b_id];
	DWORD e_id = AskEntryId(block);
	MT_ENTRY *entry = GetEntry(block, e_id);
	TCHAR text[MAX_TEXT_LEN];
	_tprintf(TEXT("You selected:\n%u:%u: "), b_id, e_id);
	PrintText(entry);
	LocalFree(entry->tText);
	_tprintf(TEXT("New Text: "));
	ReadLine(text, MAX_TEXT_LEN);
	entry->unicode = UNI;
	entry->tText = CopyStr(text);
	_tprintf(TEXT("Set text to:\n%u:%u: "), b_id, e_id);
	PrintText(entry);
}

static void AddMessage(MT_DATA *data) {
	_tprintf(TEXT("Not implemented yet\n"));
}

static void AddBlock(MT_DATA *data) {
	_tprintf(TEXT("Not implemented yet\n"));
}

static void DeleteMessage(MT_DATA *data) {
	_tprintf(TEXT("Not implemented yet\n"));
}

static void DeleteBlock(MT_DATA *data) {
	_tprintf(TEXT("Not implemented yet\n"));
}

static void SetIdBase(MT_DATA *data) {
	DWORD b_id = AskBlockId(data), id;
	MT_BLOCK *block = data->blocks[b_id];
	_tprintf(TEXT("Current ID base is %u so message IDs range from %u to %u\n"), block->lowId, block->lowId, block->highId);
	_tprintf(TEXT("New ID Base: "));
	_tscanf(TEXT("%u"), &id);
	ReadChar(); // clears the new line character
	block->lowId = id;
	UpdateEntryIds(block);
	_tprintf(TEXT("New message ID range: %u to %u\n"), block->lowId, block->highId);
}

#include "mingw-unicode.c"
int _tmain(int argc, _TCHAR* argv[]) {
	MT_DATA *data;
	BOOL saved = TRUE;

	_tprintf(TEXT("MsgTableEditor Copyright (C) 2010  Jeffrey Bush <jeff@coderforlife.com>\n"));
	_tprintf(TEXT("This program comes with ABSOLUTELY NO WARRANTY;\n"));
	_tprintf(TEXT("This is free software, and you are welcome to redistribute it\n"));
	_tprintf(TEXT("under certain conditions;\n"));
	_tprintf(TEXT("See http://www.gnu.org/licenses/gpl.html for more details.\n\n"));

	if (argc > 2) {
		_tprintf(TEXT("Usage: [file.msg]\n"));
		_tprintf(TEXT("Note: this is simply the extracted message table, not a PE file with a message table\n"));
		return 1;
	}

	if (argc == 1) {
		data = CreateBlankMessageTable();
	} else {
		data = ReadMessageTable(CopyStr(argv[1]));
		if (!data) { data = CreateBlankMessageTable(); }
	}

	DisplayMessageTable(data);

	while (TRUE) {
		int c;
		_tprintf(TEXT("\nCommand (H for help): "));
		c = toupper(ReadChar());
		if (c == EOF || c == 'Q') { if (c == EOF || saved || AskToSave(data)) { break; } }
		else if (c == 'H' || c == '?') { _tprintf(TEXT("  Q Quit\n  H Help\n  D Display Message Table\n\n  S Save\n  A Save As\n  L Load\n  N New File\n\n  R Replace Message\n  M Add Message\n  B Add Block\n  X Delete Message\n  W Delete Block\n  I Set ID Base\n")); }
		else if (c == 'D') { DisplayMessageTable(data); }
		else if (c == 'R') { ReplaceMessage(data);	saved = FALSE; }
		else if (c == 'M') { AddMessage(data);		saved = FALSE; }
		else if (c == 'B') { AddBlock(data);		saved = FALSE; }
		else if (c == 'X') { DeleteMessage(data);	saved = FALSE; }
		else if (c == 'W') { DeleteBlock(data);		saved = FALSE; }
		else if (c == 'I') { SetIdBase(data);		saved = FALSE; }
		else if (c == 'S') { SaveMessageTable(data, NULL); saved = TRUE; }
		else if (c == 'A') {
			TCHAR file[MAX_PATH];
			_tprintf(TEXT("Save As Filename: "));
			ReadLine(file, MAX_PATH);
			SaveMessageTable(data, file);
			saved = TRUE;
		} else if (c == 'L') {
			MT_DATA *data2;
			TCHAR file[MAX_PATH];
			if (!saved) { saved = AskToSave(data); }
			if (saved) {
				_tprintf(TEXT("Load Filename: "));
				ReadLine(file, MAX_PATH);
				data2 = ReadMessageTable(CopyStr(file));
				if (data2) {
					DestroyData(data);
					data = data2;
					saved = TRUE;
				}
			}
		} else if (c == 'N') {
			if (!saved) { saved = AskToSave(data); }
			if (saved) {
				DestroyData(data);
				data = CreateBlankMessageTable();
				saved = TRUE;
			}
		} else {
			_tprintf(TEXT("Command not understood: %c\n"), (char)c);
		}
	}

	DestroyData(data);

	return 0;
}
